# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import os

import numpy as np
import pandas
import six

def __matlab_round(val):
    if six.PY2:
        return round(val)
    else:
        import decimal
        return int(decimal.Decimal(val).quantize(decimal.Decimal('1'), 
    rounding=decimal.ROUND_HALF_UP))

def read_rigol_csv(filepath):
    """Reads the csv file saved from the Rigol digital storage oscilloscope.
    
    Arguments:
        filepath: A string containing the path to the csv file

    Returns:
        A tuple :code:`(X, time, CH1, CH2)` where each element is a 1D numpy 
        array containing an array of incrementing integers, an array of time 
        points (x-axis of the oscilloscope), and the y-values for channels 1
        and 2 respectively.
    """

    # read the data (skip the second line, and only use first 3 columns)
    df = pandas.read_csv(filepath, skiprows=[1], usecols=[0,1,2])

    X = df[df.columns[0]].values
    CH1 = df[df.columns[1]].values
    CH2 = df[df.columns[2]].values

    # now read the information poorly encoded in the header
    df = pandas.read_csv(filepath, nrows=1, usecols=[3,4])
    
    dt = df[df.columns[1]].values[0]
    time = X*dt

    return X, time, CH1, CH2

def extract_frequency(michelson_data, MHz_per_fringe, num_dc_terms_to_remove=1):
    """Returns a frequency calibration determined from Michelson interferometer data
    
    Performs a Hilbert transform (minus the DC term) on the Michelson
    interferometer data and scales the resultant frequency array by the
    distance (in MHz) between constructive interference fringes.
    
    Arguments:
        michelson_data: A 1D numpy array containing the oscilloscope data 
                        from the michelson interferometer

        MHz_per_fringe: The distance (in MHz) between constructive interference
                        fringes in the :code:`michelson_data` calculated from
                        the path length difference of the Michelson 
                        interferometer arms.

    Keyword Arguments:
        num_dc_terms_to_remove: The number of low frequency components to remove
                                after performing the Fourier transform. The 
                                default is 1 (remove the DC term only). Higher 
                                integers remove low order frequency components
    
    Returns:
        A 1D numpy array (with the same length as :code:`michelson_data`)
        containing a frequency calibration (in MHz) for the saturated absorption 
        spectra. Use this array as the x-axis for plotting and fitting.
    """
    ffty = np.fft.fft(michelson_data)
    subffty = np.zeros(michelson_data.shape, dtype=np.complex128)
    subffty[__matlab_round(num_dc_terms_to_remove):__matlab_round(len(michelson_data)/2)+1] = ffty[__matlab_round(num_dc_terms_to_remove):__matlab_round(len(michelson_data)/2)+1]
    back = np.fft.ifft(subffty)
    return np.unwrap(np.angle(back))/(2*np.pi)*MHz_per_fringe, back
    


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

def test_k_trap_theory():
    """Basic test of PHS3000 optical tweezers function k_trap_theory"""

    ### Get results ###
    from monashspa.PHS3000 import optical_tweezers as op

    result = op.trap_k_theory(1.5e-6, 3.2e-7, (1.58**2/1.33**2-1), 3, 5251802272)
    expected_result = 2.6373474496644467e-06
    precision = 1e-12

    success = True
    if np.abs(result-expected_result) > precision:
        success = False

    return success

def test_k_trap_theory_array():
    """Test of PHS3000 optical tweezers function k_trap_theory with array input for r"""

    ### Get results ###
    from monashspa.PHS3000 import optical_tweezers as op

    r = np.array([1.5e-6, 2e-6, 3.1e-6])
    result = op.trap_k_theory(r, 3.2e-7, (1.58**2/1.33**2-1), 3, 5251802272)
    expected_result = np.array([2.6373474496644467e-06, 7.146758330328241e-07, 2.1117891093096543e-08])
    precision = 1e-12

    success = expected_result.shape == result.shape
    if not np.all(np.isclose(expected_result, result, atol=precision)):
        success = False

    return success

def test_k_trap_theory_array_multiple():
    """Test of PHS3000 optical tweezers function k_trap_theory with array input for r and w"""

    ### Get results ###
    from monashspa.PHS3000 import optical_tweezers as op

    r = np.array([1.5e-6, 2e-6, 3.1e-6])
    w = np.array([3.2e-7, 2e-7, 8e-7])
    result = op.trap_k_theory(r, w, (1.58**2/1.33**2-1), 3, 5251802272)
    expected_result = np.array([2.6373474496644467e-06, 9.069896532126502e-09, 1.2885130290016302e-05])
    precision = 1e-12

    success = expected_result.shape == result.shape
    if not np.all(np.isclose(expected_result, result, atol=precision)):
        success = False

    return success

def do_tests():
    tests = [test_k_trap_theory, test_k_trap_theory_array, test_k_trap_theory_array_multiple]
    failed_tests = []

    for testfn in tests:
        result = testfn()
        print('Result of "{test_name}" was {result}'.format(test_name=testfn.__doc__, result='success' if result else 'failure'))

        if not result:
            failed_tests.append(testfn)

    if failed_tests:
        print('')
        print('There were {num_failures:d} failed fitting tests'.format(num_failures=len(failed_tests)))

    return failed_tests

if __name__ == "__main__":
    do_tests()


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

__attenuator_calibration = np.array([
    (0,  0.412,  0.412,  np.nan),
    (2,  0.248,  0.278,  12.14),
    (4,  0.194,  0.231,  11.75),
    (6,  0.143,  0.198,  11.47),
    (8,  0.109,  0.165,  11.22),
    (10, 0.079,  0.136,  11.02),
    (12, 0.054,  0.115,  10.84),
    (14, 0.030,  0.094,  10.66),
    (16, 0.010,  0.074,  10.50),
    (18, np.nan, 0.054,  10.33),
    (20, np.nan, 0.036,  10.17),
    (25, np.nan, np.nan, 9.87),
    (30, np.nan, np.nan, 9.27),
])

__micrometer_1_attenuation = __attenuator_calibration[:,[0,1]]
#: The calibration data for attenuator #1.
#: Here, the first column corresponds to the attentuation at 9.15GHz in dB
#: and the second column corresponds to the micrometer position in inches
micrometer_1_attenuation = __micrometer_1_attenuation[~np.isnan(__micrometer_1_attenuation[:,1])]

__micrometer_2_and_3_attenuation = __attenuator_calibration[:,[0,2]]
#: The calibration data for attenuator #2 and #3.
#: Here, the first column corresponds to the attentuation at 9.15GHz in dB
#: and the second column corresponds to the micrometer position in inches
micrometer_2_and_3_attenuation = __micrometer_2_and_3_attenuation[~np.isnan(__micrometer_2_and_3_attenuation[:,1])]

__micrometer_7925_attenuation = __attenuator_calibration[:,[0,3]]
#: The calibration data for attenuator #7925.
#: Here, the first column corresponds to the attentuation at 9.15GHz in dB
#: and the second column corresponds to the micrometer position in nm
micrometer_7925_attenuation = __micrometer_7925_attenuation[~np.isnan(__micrometer_7925_attenuation[:,1])]

__wavemeter_calibration = np.array([
    (7.7024, 24.161, 24.768, 24.215),
    (7.800,  22.675, 23.216, 22.675),
    (7.9,    21.175, 21.788, 21.255),
    (8.0,    19.895, 20.52,  19.97),
    (8.1,    18.80,  19.31,  18.81),
    (8.2,    17.75,  18.29,  17.75),
    (8.3,    16.79,  17.32,  16.79),
    (8.4,    15.87,  16.42,  15.91),
    (8.5,    15.05,  15.60,  15.06),
    (8.6,    14.30,  14.83,  14.36),
    (8.7,    13.64,  14.20,  13.64),
    (8.8,    12.92,  13.45,  12.98),
    (8.9,    12.39,  12.83,  12.35),
    (9.0,    11.70,  12.24,  11.77),
    (9.1,    11.15,  11.69,  11.21),
    (9.2,    10.63,  11.17,  10.69),
    (9.3,    10.14,  10.67,  10.20),
    (9.4,    9.66,   10.20,  9.73),
    (9.5,    9.22,   9.75,   9.29),
    (9.6,    8.80,   9.33,   8.87),
    (9.7,    8.46,   8.93,   8.46),
    (9.8,    8.08,   8.55,   8.08),
    (9.9,    7.65,   8.18,   7.71),
    (10.0,   7.31,   7.84,   7.36),
    (10.1,   6.97,   7.50,   7.03),
    (10.2,   6.65,   7.18,   6.71),
    (10.3,   6.33,   6.86,   6.39),
    (10.4,   6.03,   6.57,   6.09),
    (10.5,   5.75,   6.28,   5.81),
    (10.6,   5.47,   6.00,   5.52),
    (10.7,   5.20,   5.73,   5.26),
    (10.8,   4.94,   5.47,   5.00),
    (10.9,   4.69,   5.22,   4.75),
    (11.0,   4.45,   4.98,   4.50),
    (11.1,   4.22,   4.74,   4.28),
    (11.2,   4.00,   4.52,   4.06),
    (11.3,   3.77,   4.29,   3.84),
    (11.4,   3.56,   4.09,   3.62),
    (11.5,   3.37,   3.87,   3.42),
    (11.6,   3.17,   3.69,   3.22),
    (11.7,   2.98,   3.51,   3.03),
    (11.8,   2.80,   3.33,   2.85),
    (11.9,   2.62,   3.14,   2.66),
    (12.0,   2.44,   2.96,   2.49),
])

#: The calibration data for wavemeter S/N 4091 as a numpy array
#: Here, the first column corresponds to the frequency in GHz
#: and the second column corresponds to the wavemeter position
wavemeter_4091 = __wavemeter_calibration[:,[0,1]]

#: The calibration data for wavemeter S/N 4930 as a numpy array
#: Here, the first column corresponds to the frequency in GHz
#: and the second column corresponds to the wavemeter position
wavemeter_4930 = __wavemeter_calibration[:,[0,2]]

#: The calibration data for wavemeter S/N 4929 as a numpy array
#: Here, the first column corresponds to the frequency in GHz
#: and the second column corresponds to the wavemeter position
wavemeter_4929 = __wavemeter_calibration[:,[0,3]]


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

from .optical_tweezers import trap_k_theory as __trap_k_theory

def model_1(r, w, I):
    # scale parameters
    r = r*1e-3
    w = w*1e-4
    I = I*1e11
    
    # set parameters we are not exposing
    eccentricity = 1.1
    alpha = 1.25

    return __trap_k_theory(r, w, alpha, eccentricity, I)


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import random

import lmfit
from lmfit.models import LinearModel
import numpy as np
# from warnings import warn

class MonashSPAFittingException(Exception):
    pass

def model_fit(model, parameters, x, y, u_y=None, **kwargs):
    """A wrapper for fitting to an arbitrary model using lmfit.

    This function automatically inverts the array of standard errors
    to weights (which lmfit expects) and disables the scaling of the 
    covariant matrix is the array of standard errors are provided.

    Note: Any additional keyword arguments passed to this function
          will be passed directly to :code:`model.fit`.

    Arguments:
        model: a reference to a `lmfit`_ model.

        parameters: a reference to a :py:class:`lmfit.parameters.Parameters`
                    object for your model
        
        x: A 1D numpy array of x data points

        y: A 1D numpy array of y data points

    Keyword Arguments:
        u_y: An optional argument for providing a 1D numpy array of uncertainty 
             values for the y data points
    
    Returns:
        A :py:class:`lmfit.model.ModelResult` object from the `lmfit`_ Python library

    .. _`lmfit`: https://lmfit.github.io/lmfit-py/
    
    """
    # invert u_y because lmfit wants weights not sigma
    if u_y is not None:
        if not isinstance(u_y, np.ndarray):
            u_y = np.array(u_y)
        u_y = 1.0/u_y

        if 'scale_covar' not in kwargs:
            kwargs['scale_covar'] = False

    # if 'nan_policy' not in kwargs:
    #     kwargs['nan_policy'] = 'omit'

    #     # warn if there are any nans!
    #     to_check = [('x', x), ('y', y)]
    #     if u_y is not None:
    #         to_check.append(('u_y',u_y))
    #     for name, arr in to_check:
    #         if np.isnan(arr).any():
    #             warn('The {name} array contains at least one NaN. These data points will be ignored when performing the fit. This may cause problems when plotting the line of best fit (you will need to remove the corresponding point in all arrays).'.format(name=name))
    
    # find number of independent vars that are not in kwargs already
    missing_vars = []
    for var in model.independent_vars:
        if var not in kwargs:
            missing_vars.append(var)
    if len(missing_vars) > 1:
        raise MonashSPAFittingException('You have not passed in all of your independent variables as keyword arguments')
    elif len(missing_vars) == 0:
        x_included = False
        for var in model.independent_vars:
            if kwargs[var] == x:
                x_included = True
                break
        if not x_included:
            raise MonashSPAFittingException('You have passed in a value for the argument "x" but it is apparently not an independent arg of your model.')
    elif len(missing_vars) == 1:
        kwargs[missing_vars[0]] = x 

    fit_result = model.fit(y, parameters, weights=u_y, **kwargs)

    return fit_result

def linear_fit(x, y, u_y=None, slope_guess=None, intercept_guess=None):
    """ General purpose linear fit function.

    This function takes your x and y data (as numpy arrays) and returns a
    :py:class:`lmfit.model.ModelResult` object from the `lmfit`_ Python library.
    It attempts to fit your data to a model define by:
        :math:`y=mx+c`
    where :math:`m = slope` and :math:`c = intercept`.
    If guesses for the slope and intercept are not explicitly provided when
    calling this function, they will be inferred from the provided data arrays.

    Arguments:
        x: A 1D numpy array of x data points

        y: A 1D numpy array of y data points

    Keyword Arguments:
        u_y: An optional argument for providing a 1D numpy array of uncertainty 
             values for the y data points

        slope_guess: An optional argument for providing an initial guess for the
                     value of the slope parameter

        intercept_guess: An optional argument for providing an initial guess for the
                         value of the intercept parameter

    Returns:
        A :py:class:`lmfit.model.ModelResult` object from the `lmfit`_ Python library

    .. _`lmfit`: https://lmfit.github.io/lmfit-py/

    """
    # Create Model
    model = LinearModel()
    guess_kwargs = {}
    # Create parameter guesses
    if slope_guess is not None:
        guess_kwargs['slope'] = slope_guess
    if intercept_guess is not None:
        guess_kwargs['intercept'] = intercept_guess

    initial_parameters = model.guess(y, x=x, **guess_kwargs)

    fit_result = model_fit(model, initial_parameters, x, y, u_y)

    if not fit_result.success:
        raise MonashSPAFittingException("The call to 'linear_fit(...)' failed. Perhaps try specifying a good guess for the 'gradient_guess' and/or 'intercept_guess' keyword arguments? The error message returned by the fitting algorithm was: {error}".format(error = fit_result.message))

    return fit_result

def get_fit_parameters(fit_result):
    """ Returns the parameters from a fit result as a dictionary.

    This function takes a :py:class:`lmfit.model.ModelResult` object from the
    `lmfit`_ Python library and extracts the parameters of the fit along with
    their uncertainties. These are returned to you in a Python dictionary
    format. The format of the dictionary depends on the model used to perform
    the fit. For example, a linear fit would result in the following 
    dictionary:
        .. code-block:: python

            parameters = {
                'slope': <value>,
                'u_slope': <value>,
                'intercept': <value>,
                'u_intercept': <value>,
            }


    The parameter names always match those of the lmfit model, and the
    uncertainties are always the identical parameter name prefixed with
    :code:`u_`.

    Arguments:
        fit_result: A :py:class:`lmfit.model.ModelResult` object from the
                    `lmfit`_ Python library.

    Returns:
        A dictionary containing the fit parameters and their associated 
        uncertainties.

    .. _`lmfit`: https://lmfit.github.io/lmfit-py/

    """
    results = {}
    for param_name, param in fit_result.params.items(): 
        results[param_name] = param.value
        try:
            results['u_'+param_name] = param.stderr
        except Exception:
            results['u_'+param_name] = np.nan
    return results

# From https://bitbucket.org/labscript_suite/runmanager
# Code written by Philip Starkey and he has relicensed and contributed
# it as part of this Library
class __TraceDictionary(dict):

    def __init__(self, *args, **kwargs):
        self.trace_data = None
        dict.__init__(self, *args, **kwargs)

    def start_trace(self):
        self.trace_data = []

    def __getitem__(self, key):
        # only trace things that don't exist
        try:
            dict.__getitem__(self, key)
        except Exception:
            if self.trace_data is not None:
                if key not in self.trace_data:
                    self.trace_data.append(key)
        return dict.__getitem__(self, key)
    def stop_trace(self):
        trace_data = self.trace_data
        self.trace_data = None
        return trace_data

def make_lmfit_model(expression, independent_vars=None, **kwargs):
    """A convenience function for creating a lmfit Model from an equation in a string

    This function takes an expression containing the right hand side of an
    equation you wish to use as your fitting model, and generates a
    :py:class:`lmfit.model.Model` object from the `lmfit`_ Python library

    For example, the expression :code:`"m*x+c"` would create a model that
    would fit to linear data modelled by the equation :math:`y=m*x+c`.
    Note that the expression does not contain the :code:`y` or :code:`=`
    symbols 
    
    The independent variable is assumed to be :code:`x` unless otherwise
    specified. All other variables are assumed to be parameters you wish
    the fitting routine to optimise.

    Standard numpy and scipy.special functions are also available for use 
    in your expression.
    For example, this is also a valid expression: :code:`"sin(x)+c"`.

    The expression must always be valid Python code, and must be able to 
    be evaluated with every parameter set to a random floating point number.

    Note: Additional keyword arguments are passed directly to 
    :py:class:`lmfit.model.Model`.

    Arguments:
        expression: A string containing the right-hand-side of the equation
                    you wish to model (assumes the left hand side is equal
                    to "y").

    Keyword Arguments:
        independent_vars: a list of independent variable names that should
                          not be varied by lmfit. If set to :code:`None`
                          (the default) it assumes that the independent
                          variables is just :code:`["x"]`

    Returns:
        A :py:class:`lmfit.model.Model` object to be used for fitting with
        the lmfit library.

    .. _`lmfit`: https://lmfit.github.io/lmfit-py/
    
    """

    # assume "x" if not specified
    if independent_vars is None:
        independent_vars = ["x"]
    # ensure it's a list!
    elif not isinstance(independent_vars, list):
        independent_vars = list(independent_vars)

    # detect the parameter names in the expression if they are not provided
    sandbox = __TraceDictionary()
    # provide some basic numpy functionality
    __model_sandbox_imports(sandbox)

    # keep a list of parameters we should randomise every iteration 
    # (to prevent things like divide by 0 exceptions!)
    params_to_randomise = []
    params_to_randomise.extend(independent_vars)

    keep_trying = 5
    success = False
    while not success:
        # set random values for each parameter we know about so far
        for param in params_to_randomise:
            r = random.random()*10
            if random.random() < 0.05:
                r += random.random()*1000
            sandbox[param] = r

        # attempt to evaluate the expression
        sandbox.start_trace()
        try:
            code = compile(expression, '<string>', 'eval')
            eval(code, sandbox)
        except TypeError:
            params = sandbox.stop_trace()

            problem_param = None
            if params_to_randomise is not None:
                if params_to_randomise[-1] not in independent_vars:
                    problem_param = params_to_randomise[-1]
            if problem_param is None:
                for param in params:
                    if param not in params_to_randomise:
                        problem_param = param
            if problem_param is None:
                problem_param = "[Could not determine the parameter name]"

            raise RuntimeError('Error occurred while evaluating the model function. The problem is likely with the use of "{var}" which is either an unknown function or a parameter that cannot be set to a floating point number.'.format(var=problem_param))
        except Exception:
            # An exception was raised! This is either because:
            #   We found a new parameter name (great!)
            #   The expression has an error (Bad!)
            #
            # So we get the params found in this evaluation, and
            # see if any are new
            params = sandbox.stop_trace()
            original_length = len(params_to_randomise)
            for param in params:
                if param not in params_to_randomise:
                    params_to_randomise.append(param)

            # decide whether we keep going or not depending on whether
            # we found a new parameter
            if len(params_to_randomise) == original_length:
                keep_trying -= 1
            else:
                keep_trying += 1

            if keep_trying <= 0:
                raise
            else:
                continue
        
        # if we get to here, everything evaluated!
        success = True
        params = sandbox.stop_trace()
        for param in params:
            if param not in params_to_randomise:
                params_to_randomise.append(param)
    
    # params_to_randomise now contains everything we need!
    code_str = "def custom_model_function("
    for i, param in enumerate(params_to_randomise):
        code_str += param
        if i < len(params_to_randomise)-1:
            code_str += ", "
    code_str += "): return {expression}".format(expression=expression)

    # create fresh sandbox with standard imports
    sandbox = {}
    __model_sandbox_imports(sandbox)
    # compile our function code
    code = compile(code_str, 'model.py', 'exec')
    # execute it in the sandbox
    exec(code, sandbox, sandbox)
    # extract the model function and create the model
    model_fn = sandbox['custom_model_function']
    model = lmfit.models.Model(model_fn, independent_vars=independent_vars, **kwargs)
    return model

def __model_sandbox_imports(sandbox):
    # scipy functions
    exec('from scipy.special import *', sandbox, sandbox)
    exec('import scipy.special', sandbox, sandbox)
    # numpy functions (will overwrite some scipy in global namespace)
    # but the scipy ones will also be available through scipy.special.<function>
    exec('from numpy import *', sandbox, sandbox)
    exec('import numpy as np', sandbox, sandbox)


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pandas

def read_data(filepath):
    """Reads the csv file produced by the thermoelectricity software

    Arguments:
        filepath: The path to the .csv file

    Returns:
        A 2D numpy array containing the data from the file, with columns
        date-time, delta_t from row 0, T_Cold, T_Hot, T_Case, I_Heat, V_Heat,
        I_Pelt, V_Pelt, respectively.

    """

    df = pandas.read_csv(filepath, parse_dates={'Datetime':[0,1]})

    # convert to numpy array
    arr = df.to_numpy()

    # calculate timedeltas
    time_delta = np.zeros((arr.shape[0],))

    # vectorise the conversion to total seconds
    f = np.vectorize(lambda x: x.total_seconds())

    # calculate the timedelta column
    time_delta[1:] = f(arr[1:,0]-arr[0,0])

    # insert the timedelta column at column 1
    arr = np.insert(arr, 1, time_delta, axis=1)

    return arr


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pandas

def read_data(filepath):
    """Reads the csv file produced by the betaray online experiment

    Arguments:
        filepath: The path to the .csv file

    Returns:
        A 2D numpy array containing the data from the file, with columns
        Acquisition time, Run mode, Vacuum pressure, Shutter Position, Duration (s), Count, Lens coil current (A), u(Lens coil current) (A), Bias coil current (A), u(Bias coil current) (A), Magnetometer x position, Magnetometer y position, Magnetometer x-axis field strength (µT), u(Magnetometer x-axis field strength) (µT), Magnetometer y-axis field strength (µT), u(Magnetometer y-axis field strength) (µT), Magnetometer z-axis field strength (µT), and u(Magnetometer z-axis field strength) (µT) respectively.

    """

    df = pandas.read_csv(filepath, parse_dates=[0])

    # convert to numpy array
    arr = df.to_numpy()

    return arr

#: The modified Fermi function, :math:`G`, for :math:`Z=55`.
#: Here, the first column corresponds to momentum, :math:`p`, in relativistic 
#: units of :math:`m_0 c^2` and the second column cooresponds to the value of
#: the modified Fermi function :math:`G`.
modified_fermi_function_data = np.array([
    [0.0, 6.591],
    [0.1, 6.582],
    [0.2, 6.552],
    [0.3, 6.506],
    [0.4, 6.448],
    [0.5, 6.387],
    [0.6, 6.329],
    [0.7, 6.275],
    [0.8, 6.224],
    [0.9, 6.177],
    [1.0, 6.132],
    [1.2, 6.046],
    [1.4, 5.964],
    [1.6, 5.886],
    [1.8, 5.812],
    [2.0, 5.742],
    [2.2, 5.675],
    [2.4, 5.612],
    [2.6, 5.553],
    [2.8, 5.496],
    [3.0, 5.443],
])


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import os

import numpy as np
import pandas
import matplotlib.pyplot as plt
import lmfit
from lmfit.models import QuadraticModel as __QuadraticModel
import scipy.special


from monashspa.PHS3000 import linear_fit as __linear_fit
from monashspa.common.fitting import MonashSPAFittingException as __MonashSPAFittingException

def trap_k_theory(r, w, alpha, eccentricity, I):
    r"""Calculates the theoretical spring constant (:math:`k`) for an optical tweezers trap for specified parameters

    Calculates using the equation:
        :math:`k=\alpha\,I_0\,\omega\,\frac{2\,\pi\,\epsilon^3}{c\,\xi^3}\left(\sqrt{\frac{\pi}{2}}\,\left(\left(\frac{\xi\,a}{\epsilon}\right)^2-1\right)\,\mathrm{exp}\left[-\frac{a^2}{2}\right]\,\mathrm{erf}\left[\frac{\xi\,a}{\sqrt{2}\,\epsilon}\right]+\frac{\xi\,a}{\epsilon}\,\mathrm{exp}\left[-\frac{a^2}{2\,\epsilon^2}\right]\right)`
    from `Bechhoefer 2002`_.

    If the input arguments are numpy arrays, then the output will also be an
    array of the appropriate dimension. Otherwise a single number will be 
    returned.

    Arguments:
        r: Sphere radius (m)
        w: The :math:`1/e^2` radius (beam waist) of the trapping beam (m)
        alpha: :math:`n_p^2/n_0^2 - 1`, where :math:`n_p` is the refractive 
               index of the microsphere and :math:`n_0` is the refractive index
               of water
        eccentricity: The eccentiricty of the trapping beam
        I: Trapping beam intensity (W/m^2)

    Returns:
        k: The theoretical spring constant (:math:`k`)

    .. _`Bechhoefer 2002`: http://scitation.aip.org.ezproxy.lib.monash.edu.au/content/aapt/journal/ajp/70/4/10.1119/1.1445403

    """
    
    # calculate a
    a = r/w

    # Make eccentricity complex so that numpy can handle the sqrt
    eccentricity = eccentricity + 0j
    xi = np.sqrt(1-eccentricity**2)
    
    # define speed of light
    c = 299792458

    # calculate k
    k = alpha/c*I*w*2*np.pi*eccentricity**3/xi**3*(
            np.sqrt(np.pi/2)*((xi*a/eccentricity)**2-1)*np.exp(-a**2/2)*scipy.special.erf(xi*a/(np.sqrt(2)*eccentricity)) 
            + (xi*a/eccentricity)*np.exp(-a**2/(2*eccentricity**2))
        )

    # if the imaginary component is 0 (which it should be), then return only
    # the real component. Otherwise, return the whole complex number
    if isinstance(k, np.ndarray):
        return np.real_if_close(k)
    else:
        return k.real if k.imag == 0 else k

def ps_load(filepath):
    """Imports the power spectrum data from optical tweezers file

    Arguments:
        filepath: The path to the .lvm file produced by the optical tweezers
                  acquisition software

    Returns:
        A tuple :code:`(f, psx, psy)` where :code:`f` is a 1D numpy array
        containing the frequency values associated with the power spectrums
        in :code:`psx` and :code:`psy` (which are also both 1D numpy arrays).

    """
    # TODO: consider replacing with our own csv reading wrapper
    df = pandas.read_csv(filepath, skiprows=22, sep='\t')
    f = df[df.columns[8]].values
    psx = df[df.columns[9]].values
    psy = df[df.columns[11]].values

    # strip out NaNs (as these columns are shorter than the others)
    f = f[~np.isnan(f)]
    psx = psx[~np.isnan(psx)]
    psy = psy[~np.isnan(psy)]

    return f, psx, psy

def cf_linearised(f, ps, initial_fc, call_show = True):
    r"""Finds the corner frequency value for a lorentzian power spectrum

    Finds the corner frequency (:math:`f_c`) of a power spectrum of the form:
        :math:`y=\frac{a}{(f_c^2+f^2)}`
    by transforming into the logarithmic domain and determining when the 
    spectrum is linearised.

    Arguments:
        f: A 1D numpy array containing the frequency values associated with 
           the power spectrum
        ps: A 1D numpy array containing the power spectrum data (must be the
            same length as :code:`f`)
        initial_fc: An initial guess for the corner frequency

    Keyword Arguments:
        call_show: Whether to call :code:`matplotlib.pyplot.show()` at the end
                   of the function (prior to returning). Defaults to :code:`True`.
                   Set this to :code:`False` if you are not using Spyder/IPython
                   and wish your entire script to complete before showing any 
                   plots. Note, you will need to explicitly call 
                   :code:`matplotlib.pyplot.show()` if you set this to :code:`False`.
    
    Returns:
        The best estimate for the corner frequency :math:`f_c`.

    """
    y = np.log(ps)

    plt.figure()
    plt.loglog(f,ps)
    plt.title('Power spectrum')
    plt.xlabel('Frequency (Hz)')
    plt.ylabel(r'Power Spectral Density ($V^2$/Hz)')
    
    # remove first 2 terms
    y = y[2:]
    f = f[2:]
    ps = ps[2:]

    params = lmfit.Parameters()
    params.add('a', y[0])
    params.add('fc', initial_fc)
    result = lmfit.minimize(__cf_linearised_search, params, method='nelder', args=(f,y))
    fc = result.params['fc'].value

    plt.figure()
    plt.semilogy(np.log(fc**2+f**2), ps)
    plt.title('Linearised Power Spectrum Analysis')
    plt.xlabel(r'$f_0^2+f^2$ ($Hz^2$)')
    plt.ylabel(r'Power Spectral Density ($V^2$/Hz)')

    if call_show:
        plt.show()

    return fc

def __cf_linearised_search(params, f, y):
    parvals = params.valuesdict()
    a = parvals['a']
    fc = parvals['fc']

    out = a - np.log(fc**2+f**2)

    # do linear fit
    linear_result = __linear_fit(out, y)

    # do quadratic fit
    quadratic_result = __quadratic_fit(out, y)

    e = linear_result.best_fit - quadratic_result.best_fit
    return np.transpose(e)@e

def __quadratic_fit(x, y):
    """Special purpose quadratic fit for optical tweezers

    Arguments:
        x: A 1D numpy array of x data points

        y: A 1D numpy array of y data points

    Returns:
        A :py:class:`lmfit.model.ModelResult` object from the `lmfit`_ Python library

    .. _`lmfit`: https://lmfit.github.io/lmfit-py/

    """
    # Create Model
    model = __QuadraticModel()
    initial_parameters = model.guess(y, x=x)
    fit_result = model.fit(y, initial_parameters, x=x)

    if not fit_result.success:
        raise __MonashSPAFittingException("Failed to perform the quadratic fit step. The error message returned by the fitting algorithm was: {error}".format(error = fit_result.message))

    return fit_result


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division

import os

import numpy as np
import pandas
import matplotlib.pyplot as plt
from piradon import iradon as __iradon

def pet_rebuild(filepath, filter_name=None, npoints=None, call_show=True):
    """Perform inverse radon transform on acquired PET data and plots results

    Arguments:
        filepath: A string containing the path to the txt file containing the 
                  PET data
    
    Keyword Arguments:
        filter_name: A string containing the name of the filter to use during
                     reconstruction. Defaults to :code:`None` (no
                     reconstruction). Options are:
                        None: don't reconstruct
                        'none': Reconstruct with no filter
                        'ramp': Reconstruct using the Ram-Lak filter
                        'Shepp-Logan': Reconstruct using the Shepp-Logan filter
                        'cosine': Reconstruct using the cosine filter
                        'hamming': Reconstruct using the hamming filter
                        'hann': Reconstruct using the hann filter

        npoints: The number of points to reconstruct

        call_show: Whether to call :code:`matplotlib.pyplot.show()` at the end
                   of the function (prior to returning). Defaults to :code:`True`.
                   Set this to :code:`False` if you are not using Spyder/IPython
                   and wish your entire script to complete before showing any 
                   plots. Note, you will need to explicitly call 
                   :code:`matplotlib.pyplot.show()` if you set this to :code:`False`. 
    
    Returns:
        A 2D numpy array containing the coincidence counts (rows correspond to
        each linear stage position and columns to each rotation stage position)

    """

    # TODO: consider replacing with our own csv reading wrapper
    df = pandas.read_csv(filepath, skiprows=1, sep=',\t', engine='python', parse_dates=[0])

    # acquisition_time = df[df.columns[0]].values
    rotations = df[df.columns[1]].values
    positions = df[df.columns[2]].values
    coincidence_counts = df[df.columns[3]].values

    # determine number of points in each axis of the scan
    unique_angles = np.unique(rotations)
    unique_positions = np.unique(positions)

    # reshape coincident counts array
    #
    # Note: The array dimensions are swapped between MATLAB and Python.
    #       This avoids the need to transpose as you do in MATLAB because 
    #       of the different scan order during reshape between MATLAB and 
    #       Python. See: https://docs.scipy.org/doc/numpy-1.15.0/user/numpy-for-matlab-users.html#notes
    coincidence_counts.shape = (len(unique_positions), len(unique_angles))

    plt.figure()
    plt.imshow(coincidence_counts, extent=[np.min(unique_angles), np.max(unique_angles), np.max(unique_positions), np.min(unique_positions)], interpolation='none')
    plt.xlabel(r'$\theta$')
    plt.ylabel('x')
    plt.title('Sinogram of {filename}'.format(filename=os.path.basename(filepath)))

    # do the inverse transform
    if filter_name is not None:
        if npoints is None:
            raise RuntimeError('When calling pet_rebuild with a filter, you must specify the number of points to reconstruct')

        # convert string none to actual None
        if filter_name == 'none':
            filter_name = None
        inverse = __iradon(coincidence_counts, unique_angles, output_size=npoints, filter=filter_name, interpolation='linear', circle=False)

        plt.figure()
        plt.imshow(inverse, extent=[0, 1, 0, 1], interpolation='none')
        plt.title('Reconstruction of {filename}'.format(filename=os.path.basename(filepath)))

    if call_show:
        plt.show()

    return coincidence_counts
